"""Write HTTP Archive (HAR) files."""

from __future__ import annotations

import builtins
import json
from dataclasses import asdict
from datetime import datetime
from os import PathLike
from types import TracebackType
from typing import IO, Any

from ._models import (
    Browser,
    Cache,
    Content,
    Cookie,
    Creator,
    PostData,
    PostParameter,
    Record,
    Request,
    Response,
    Timings,
)
from ._version import VERSION

__all__ = [
    "open",
    "HarFile",
    "Browser",
    "Creator",
    "Cache",
    "Request",
    "Response",
    "Timings",
    "Cookie",
    "Record",
    "PostData",
    "PostParameter",
    "Content",
    "HAR_VERSION",
]

HAR_VERSION = "1.2"


class HarFile:
    _fd: IO[str]
    _creator: Creator
    _browser: Browser
    _comment: str | None
    _is_first_entry: bool = True
    _has_preable: bool = False
    closed: bool = False

    def __init__(
        self,
        fd: IO[str],
        *,
        creator: Creator | None = None,
        browser: Browser | None = None,
        comment: str | None = None,
    ):
        self._fd = fd
        self._creator = creator or Creator(name="harfile", version=VERSION)
        self._browser = browser or Browser(name="", version="")
        self._comment = comment

    @classmethod
    def open(
        cls,
        name_or_fd: str | PathLike | IO[str],
        *,
        creator: Creator | None = None,
        browser: Browser | None = None,
        comment: str | None = None,
    ) -> HarFile:
        """Open a HAR file for writing."""
        fd: IO[str]
        if isinstance(name_or_fd, (str, PathLike)):
            fd = builtins.open(name_or_fd, "w")
        else:
            fd = name_or_fd
        return cls(fd=fd, creator=creator, browser=browser, comment=comment)

    def close(self) -> None:
        """Close the HAR file."""
        if self.closed:
            return None
        self.closed = True
        if not self._has_preable:
            self._write_preamble()
            self._has_preable = True
        self._write_postscript()

    def __enter__(self) -> HarFile:
        return self

    def __exit__(
        self,
        type: type[BaseException] | None,
        value: BaseException | None,
        traceback: TracebackType | None,
    ) -> None:
        if type is None:
            self.close()
        return None

    def add_entry(
        self,
        *,
        startedDateTime: datetime,
        time: int | float,
        request: Request,
        response: Response,
        timings: Timings,
        cache: Cache | None = None,
        serverIPAddress: str | None = None,
        connection: str | None = None,
        comment: str | None = None,
    ) -> None:
        """Add an HTTP request to the HAR file."""
        if not self._has_preable:
            self._write_preamble()
            self._has_preable = True
        self._write_entry(
            startedDateTime=startedDateTime,
            time=time,
            request=request,
            response=response,
            cache=cache,
            timings=timings,
            serverIPAddress=serverIPAddress,
            connection=connection,
            comment=comment,
        )

    def _write_entry(
        self,
        *,
        startedDateTime: datetime,
        time: int | float,
        request: Request,
        response: Response,
        timings: Timings,
        cache: Cache | None = None,
        serverIPAddress: str | None = None,
        connection: str | None = None,
        comment: str | None = None,
    ) -> None:
        separator = "\n" if self._is_first_entry else ",\n"
        self._is_first_entry = False
        self._fd.write(f"{separator}            {{")
        self._fd.write(f'\n                "startedDateTime": "{startedDateTime.isoformat()}",')
        self._fd.write(f'\n                "time": {time},')
        self._fd.write(f'\n                "request": {json.dumps(asdict(request, dict_factory=_dict_factory))},')
        self._fd.write(f'\n                "response": {json.dumps(asdict(response, dict_factory=_dict_factory))},')
        self._fd.write(f'\n                "timings": {json.dumps(asdict(timings, dict_factory=_dict_factory))}')
        if cache:
            self._fd.write(f',\n                "cache": {json.dumps(asdict(cache, dict_factory=_dict_factory))}')
        if serverIPAddress:
            self._fd.write(f',\n                "serverIPAddress": {json.dumps(serverIPAddress)}')
        if connection:
            self._fd.write(f',\n                "connection": {json.dumps(connection)}')
        if comment:
            self._fd.write(f',\n                "comment": {json.dumps(comment)}')
        self._fd.write("\n            }")

    def _write_preamble(self) -> None:
        creator = f"""{{
            "name": "{self._creator.name}",
            "version": "{self._creator.version}\""""
        if self._creator.comment:
            creator += f',\n            "comment": "{self._creator.comment}"'
        creator += "\n        }"
        browser = f"""{{
            "name": "{self._browser.name}",
            "version": "{self._browser.version}\""""
        if self._browser.comment:
            browser += f',\n            "comment": "{self._browser.comment}"'
        browser += "\n        }"
        self._fd.write(f"""{{
    "log": {{
        "version": "{HAR_VERSION}",
        "creator": {creator},
        "browser": {browser}""")
        if self._comment:
            self._fd.write(f'    "comment": "{self._comment}"')
        self._fd.write(',\n        "entries": [')

    def _write_postscript(self) -> None:
        if self._is_first_entry:
            self._fd.write("]\n    }\n}")
        else:
            self._fd.write("\n        ]\n    }\n}")


def _dict_factory(value: list[tuple[str, Any]]) -> dict[str, Any]:
    return {key: value for key, value in value if value is not None}


open = HarFile.open



from __future__ import annotations

from dataclasses import dataclass, field


@dataclass
class Creator:
    # Name of the application used to export the log
    name: str
    # Version of the application used to export the log
    version: str
    comment: str | None = None


@dataclass
class Browser:
    # Name of the browser used to export the log
    name: str
    # Version of the browser used to export the log
    version: str
    comment: str | None = None


@dataclass
class Cache:
    """Info about a request coming from browser cache."""

    # State of a cache entry before the request
    beforeRequest: CacheEntry | None = None
    # State of a cache entry after the request
    afterRequest: CacheEntry | None = None
    comment: str | None = None


@dataclass
class CacheEntry:
    # Expiration time of the cache entry
    expires: str
    # The last time the cache entry was opened
    lastAccess: str
    eTag: str
    # The number of times the cache entry has been opened
    hitCount: int
    comment: str | None = None


@dataclass
class Request:
    method: str
    # Absolute URL of the request (fragments are not included)
    url: str
    # Request HTTP Version
    httpVersion: str
    cookies: list[Cookie] = field(default_factory=list)
    headers: list[Record] = field(default_factory=list)
    queryString: list[Record] = field(default_factory=list)
    postData: PostData | None = None
    # Total number of bytes from the start of the HTTP request message until (and including) the double CRLF
    # before the body
    headersSize: int = -1
    # Size of the request body (POST data payload) in bytes
    bodySize: int = -1
    comment: str | None = None


@dataclass
class Cookie:
    # The name of the cookie
    name: str
    # The cookie value
    value: str
    # The path pertaining to the cookie
    path: str | None = None
    # The host of the cookie
    domain: str | None = None
    # Cookie expiration time
    expires: str | None = None
    # Set to true if the cookie is HTTP only, false otherwise
    httpOnly: bool | None = None
    # True if the cookie was transmitted over ssl, false otherwise
    secure: bool | None = None
    comment: str | None = None


@dataclass
class Record:
    name: str
    value: str
    comment: str | None = None


@dataclass
class PostData:
    # MIME type of the posted data
    mimeType: str
    # List of posted parameters (in case of URL encoded parameters)
    params: list[PostParameter] | None = None
    # Plain text posted data
    text: str | None = None
    comment: str | None = None


@dataclass
class PostParameter:
    name: str
    value: str | None = None
    fileName: str | None = None
    contentType: str | None = None
    comment: str | None = None


@dataclass
class Content:
    size: int = 0
    # Number of bytes saved
    compression: int | float | None = None
    # MIME type of the response text (value of the Content-Type response header).
    # The charset attribute of the MIME type is included
    mimeType: str | None = None
    # Response body sent from the server or loaded from the browser cache.
    # This field is populated with textual content only.
    # The text field is either HTTP decoded text or a encoded (e.g. "base64") representation of the response body
    text: str | None = None
    encoding: str | None = None
    # Encoding used for response text field e.g "base64"
    comment: str | None = None


@dataclass
class Response:
    # Response status
    status: int
    # Response status description
    statusText: str
    # Response HTTP Version
    httpVersion: str
    cookies: list[Cookie] = field(default_factory=list)
    headers: list[Record] = field(default_factory=list)
    # Details about the response body
    content: Content = field(default_factory=Content)
    # Redirection target URL from the Location response header
    redirectURL: str = ""
    # Total number of bytes from the start of the HTTP response message
    # until (and including) the double CRLF before the body
    headersSize: int = -1
    # Size of the received response body in bytes
    bodySize: int = -1
    comment: str | None = None


@dataclass
class Timings:
    # Time required to send HTTP request to the server
    send: int | float
    # Waiting for a response from the server
    wait: int | float
    # Time required to read entire response from the server (or cache)
    receive: int | float
    # Time spent in a queue waiting for a network connection
    blocked: int | float = -1
    # DNS resolution time. The time required to resolve a host name
    dns: int | float = -1
    # Time required to create TCP connection
    connect: int | float = -1
    # Time required for SSL/TLS negotiation
    ssl: int | float = -1
    comment: str | None = None



import datetime
import io
import json
import sys

import jsonschema
import pytest
from hypothesis import HealthCheck, Phase, given, settings
from hypothesis import strategies as st

import harfile

# Derived from http://www.softwareishard.com/har/viewer/
DATETIME_PATTERN = r"^(\d{4})(-)?(\d\d)(-)?(\d\d)(T)?(\d\d)(:)?(\d\d)(:)?(\d\d)(\.\d+)?(Z|([+-])(\d\d)(:)?(\d\d))"
RECORD_SCHEMA = {
    "type": "object",
    "properties": {
        "name": {"type": "string"},
        "value": {"type": "string"},
        "comment": {"type": "string"},
    },
    "required": ["name", "value"],
    "additionalProperties": False,
}
COOKIE_SCHEMA = {
    "type": "object",
    "properties": {
        "name": {"type": "string"},
        "value": {"type": "string"},
        "path": {"type": "string"},
        "domain": {"type": "string"},
        "expires": {"type": "string"},
        "httpOnly": {"type": "boolean"},
        "secure": {"type": "boolean"},
        "comment": {"type": "string"},
    },
    "required": ["name", "value"],
    "additionalProperties": False,
}
CACHE_ENTRY_SCHEMA = {
    "type": "object",
    "properties": {
        "expires": {"type": "string"},
        "lastAccess": {"type": "string"},
        "eTag": {"type": "string"},
        "hitCount": {"type": "integer"},
        "comment": {"type": "string"},
    },
    "required": ["lastAccess", "eTag", "hitCount"],
    "additionalProperties": False,
}
HAR_SCHEMA = {
    "type": "object",
    "properties": {
        "log": {
            "type": "object",
            "properties": {
                "version": {"type": "string"},
                "creator": {
                    "type": "object",
                    "properties": {
                        "name": {"type": "string"},
                        "version": {"type": "string"},
                        "comment": {"type": "string"},
                    },
                    "required": ["name", "version"],
                    "additionalProperties": False,
                },
                "browser": {
                    "type": "object",
                    "properties": {
                        "name": {"type": "string"},
                        "version": {"type": "string"},
                        "comment": {"type": "string"},
                    },
                    "required": ["name", "version"],
                    "additionalProperties": False,
                },
                "pages": {
                    "type": "array",
                    "items": {
                        "type": "object",
                        "properties": {
                            "startedDateTime": {
                                "type": "string",
                                "format": "date-time",
                                "pattern": DATETIME_PATTERN,
                            },
                            # NOTE: The original schema requires unique `id`, but JSON Schema does not support
                            # uniqueness based on a property value.
                            "id": {"type": "string"},
                            "title": {"type": "string"},
                            "pageTimings": {
                                "type": "object",
                                "properties": {
                                    "onContentLoad": {"type": "number"},
                                    "onLoad": {"type": "number"},
                                    "comment": {"type": "string"},
                                },
                                "additionalProperties": False,
                            },
                            "comment": {"type": "string"},
                        },
                        "required": ["startedDateTime", "id", "title", "pageTimings"],
                        "additionalProperties": False,
                    },
                },
                "entries": {
                    "type": "array",
                    "items": {
                        "type": "object",
                        "properties": {
                            "pageref": {"type": "string"},
                            "startedDateTime": {
                                "type": "string",
                                "format": "date-time",
                                "pattern": DATETIME_PATTERN,
                            },
                            "time": {"type": "number"},
                            "request": {
                                "type": "object",
                                "properties": {
                                    "method": {"type": "string"},
                                    "url": {"type": "string"},
                                    "httpVersion": {"type": "string"},
                                    "cookies": {
                                        "type": "array",
                                        "items": COOKIE_SCHEMA,
                                    },
                                    "headers": {
                                        "type": "array",
                                        "items": RECORD_SCHEMA,
                                    },
                                    "queryString": {
                                        "type": "array",
                                        "items": RECORD_SCHEMA,
                                    },
                                    "postData": {
                                        "type": "object",
                                        "properties": {
                                            "mimeType": {"type": "string"},
                                            "text": {"type": "string"},
                                            "params": {
                                                "type": "array",
                                                "items": {
                                                    "type": "object",
                                                    "properties": {
                                                        "name": {"type": "string"},
                                                        "value": {"type": "string"},
                                                        "fileName": {"type": "string"},
                                                        "contentType": {"type": "string"},
                                                        "comment": {"type": "string"},
                                                    },
                                                    "required": ["name"],
                                                    "additionalProperties": False,
                                                },
                                            },
                                            "comment": {"type": "string"},
                                        },
                                        "required": ["mimeType"],
                                        "additionalProperties": False,
                                    },
                                    "headersSize": {"type": "integer"},
                                    "bodySize": {"type": "integer"},
                                    "comment": {"type": "string"},
                                },
                                "required": [
                                    "method",
                                    "url",
                                    "httpVersion",
                                    "cookies",
                                    "headers",
                                    "queryString",
                                    "headersSize",
                                    "bodySize",
                                ],
                                "additionalProperties": False,
                            },
                            "response": {
                                "type": "object",
                                "properties": {
                                    "status": {"type": "integer"},
                                    "statusText": {"type": "string"},
                                    "httpVersion": {"type": "string"},
                                    "cookies": {
                                        "type": "array",
                                        "items": COOKIE_SCHEMA,
                                    },
                                    "headers": {
                                        "type": "array",
                                        "items": RECORD_SCHEMA,
                                    },
                                    "content": {
                                        "type": "object",
                                        "properties": {
                                            "size": {"type": "integer"},
                                            "compression": {"type": "number"},
                                            "mimeType": {"type": "string"},
                                            "text": {"type": "string"},
                                            "encoding": {"type": "string"},
                                            "comment": {"type": "string"},
                                        },
                                        "required": ["size"],
                                        "additionalProperties": False,
                                    },
                                    "redirectURL": {"type": "string"},
                                    "headersSize": {"type": "integer"},
                                    "bodySize": {"type": "integer"},
                                    "comment": {"type": "string"},
                                },
                                "required": [
                                    "status",
                                    "statusText",
                                    "httpVersion",
                                    "cookies",
                                    "headers",
                                    "headersSize",
                                    "bodySize",
                                ],
                                "additionalProperties": False,
                            },
                            "cache": {
                                "type": "object",
                                "properties": {
                                    "beforeRequest": CACHE_ENTRY_SCHEMA,
                                    "afterRequest": CACHE_ENTRY_SCHEMA,
                                    "comment": {"type": "string"},
                                },
                                "additionalProperties": False,
                            },
                            "timings": {
                                "type": "object",
                                "properties": {
                                    "dns": {"type": "number"},
                                    "connect": {"type": "number"},
                                    "blocked": {"type": "number"},
                                    "send": {"type": "number"},
                                    "wait": {"type": "number"},
                                    "receive": {"type": "number"},
                                    "ssl": {"type": "number"},
                                    "comment": {"type": "string"},
                                },
                                "required": ["send", "wait", "receive"],
                                "additionalProperties": False,
                            },
                            "serverIPAddress": {"type": "string"},
                            "connection": {"type": "string"},
                            "comment": {"type": "string"},
                        },
                        "required": [
                            "startedDateTime",
                            "time",
                            "request",
                            "response",
                            "timings",
                        ],
                        "additionalProperties": False,
                    },
                },
                "comment": {"type": "string"},
            },
            "required": ["version", "creator", "browser", "entries"],
            "additionalProperties": False,
        }
    },
    "required": ["log"],
    "additionalProperties": False,
}
HAR_VALIDATOR = jsonschema.Draft7Validator(HAR_SCHEMA)


if sys.version_info < (3, 10):
    st.register_type_strategy(
        harfile.Request,
        st.builds(
            harfile.Request,
            bodySize=st.integers(),
            comment=st.text() | st.none(),
            cookies=st.lists(
                st.builds(
                    harfile.Cookie,
                    comment=st.text() | st.none(),
                    domain=st.text() | st.none(),
                    expires=st.text() | st.none(),
                    httpOnly=st.booleans() | st.none(),
                    name=st.text(),
                    path=st.text() | st.none(),
                    secure=st.booleans() | st.none(),
                    value=st.text(),
                )
            ),
            headers=st.lists(
                st.builds(
                    harfile.Record,
                    comment=st.text() | st.none(),
                    name=st.text(),
                    value=st.text(),
                )
            ),
            headersSize=st.integers(),
            httpVersion=st.text(),
            method=st.text(),
            postData=st.builds(
                harfile.PostData,
                comment=st.text() | st.none(),
                mimeType=st.text(),
                params=st.lists(
                    st.builds(
                        harfile.PostParameter,
                        comment=st.text() | st.none(),
                        contentType=st.text() | st.none(),
                        fileName=st.text() | st.none(),
                        name=st.text(),
                        value=st.text() | st.none(),
                    )
                ),
            ),
            queryString=st.lists(
                st.builds(
                    harfile.Record,
                    comment=st.text() | st.none(),
                    name=st.text(),
                    value=st.text(),
                )
            ),
            url=st.text(),
        ),
    )
    st.register_type_strategy(
        harfile.Response,
        st.builds(
            harfile.Response,
            bodySize=st.integers(),
            comment=st.text() | st.none(),
            content=st.builds(
                harfile.Content,
                comment=st.text() | st.none(),
                compression=st.floats(allow_nan=False, allow_infinity=False) | st.integers(),
                encoding=st.text() | st.none(),
                mimeType=st.text() | st.none(),
                size=st.integers(),
                text=st.text() | st.none(),
            ),
            cookies=st.lists(
                st.builds(
                    harfile.Cookie,
                    comment=st.text() | st.none(),
                    domain=st.text() | st.none(),
                    expires=st.text() | st.none(),
                    httpOnly=st.booleans() | st.none(),
                    name=st.text(),
                    path=st.text() | st.none(),
                    secure=st.booleans() | st.none(),
                    value=st.text(),
                )
            ),
            headers=st.lists(
                st.builds(
                    harfile.Record,
                    comment=st.text() | st.none(),
                    name=st.text(),
                    value=st.text(),
                )
            ),
            headersSize=st.integers(),
            httpVersion=st.text(),
            redirectURL=st.text(),
            status=st.integers(),
            statusText=st.text(),
        ),
    )
    st.register_type_strategy(
        harfile.Timings,
        st.builds(
            harfile.Timings,
            blocked=st.floats() | st.integers(),
            comment=st.text() | st.none(),
            connect=st.floats() | st.integers(),
            dns=st.floats() | st.integers(),
            receive=st.floats() | st.integers(),
            send=st.floats() | st.integers(),
            ssl=st.floats() | st.integers(),
            wait=st.floats() | st.integers(),
        ),
    )


ENTRY_STRATEGY = st.fixed_dictionaries(
    {
        "startedDateTime": st.datetimes(timezones=st.just(datetime.timezone.utc)),
        "time": st.floats(min_value=0, allow_nan=False, allow_infinity=False) | st.integers(min_value=0),
        "request": st.from_type(harfile.Request),
        "response": st.from_type(harfile.Response),
        "timings": st.from_type(harfile.Timings),
    },
    optional={
        "cache": st.from_type(harfile.Cache) | st.none(),
        "serverIPAddress": st.text() | st.none(),
        "connection": st.text() | st.none(),
        "comment": st.text() | st.none(),
    },
)


def write_har(arg, entries):
    with harfile.open(arg) as har:
        for entry in entries:
            har.add_entry(**entry)


@given(entries=st.lists(ENTRY_STRATEGY, max_size=10))
@settings(
    phases=[Phase.reuse, Phase.generate, Phase.shrink],
    suppress_health_check=[HealthCheck.function_scoped_fixture, HealthCheck.too_slow],
)
def test_write_har(entries, tmp_path):
    buffer = io.StringIO()
    write_har(buffer, entries)
    har = buffer.getvalue()
    HAR_VALIDATOR.validate(json.loads(har))
    path = tmp_path / "test.har"
    write_har(path, entries)
    with open(path) as fd:
        HAR_VALIDATOR.validate(json.load(fd))
    write_har(path, entries)
    with open(path) as fd:
        HAR_VALIDATOR.validate(json.load(fd))


def test_close():
    buffer = io.StringIO()
    with harfile.open(buffer) as har:
        har.close()


def test_exception():
    buffer = io.StringIO()
    with pytest.raises(ZeroDivisionError):
        with harfile.open(buffer):
            raise ZeroDivisionError
    assert buffer.getvalue() == ""


def test_with_comments():
    buffer = io.StringIO()
    with harfile.open(
        buffer,
        creator=harfile.Creator(name="test", version="0.1", comment="EXAMPLE-1"),
        browser=harfile.Browser(name="test", version="0.2", comment="EXAMPLE-2"),
    ):
        pass
    assert (
        buffer.getvalue()
        == """{
    "log": {
        "version": "1.2",
        "creator": {
            "name": "test",
            "version": "0.1",
            "comment": "EXAMPLE-1"
        },
        "browser": {
            "name": "test",
            "version": "0.2",
            "comment": "EXAMPLE-2"
        },
        "entries": []
    }
}"""
    )
